import pandas as pd

from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.sqlalchemy import BinaryExpression, ColumnClause
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column.values_non_null import (
    ColumnValuesNonNull,
    ColumnValuesNonNullCount,
    ColumnValuesNonNullCountResult,
    ColumnValuesNonNullResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import PANDAS_DATA_SOURCES, SPARK_DATA_SOURCES, SQL_DATA_SOURCES

STRING_COLUMN_NAME = "letter"
DATA_FRAME = pd.DataFrame(
    {
        STRING_COLUMN_NAME: ["a", None, "c", "d"],
    },
)


class TestColumnValuesNonNull:
    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_pandas(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = ColumnValuesNonNull(column=STRING_COLUMN_NAME)
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, ColumnValuesNonNullResult)
        assert isinstance(metric_result.value, pd.Series)
        expected_value = pd.Series(
            [False, True, False, False],
            name=STRING_COLUMN_NAME,
            dtype=bool,
        )
        assert metric_result.value.equals(expected_value)

    @parameterize_batch_for_data_sources(
        data_source_configs=SPARK_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_spark(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = ColumnValuesNonNull(column=STRING_COLUMN_NAME)
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, ColumnValuesNonNullResult)
        expected_value = ~(F.col(STRING_COLUMN_NAME).isNotNull())
        assert str(metric_result.value) == str(expected_value)

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_sql(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = ColumnValuesNonNull(column=STRING_COLUMN_NAME)
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, ColumnValuesNonNullResult)
        assert isinstance(metric_result.value, BinaryExpression)
        expected_value = ColumnClause(STRING_COLUMN_NAME).is_(None)
        assert metric_result.value.compare(expected_value)


class TestColumnValuesNonNullCount:
    NON_NULL_COUNT = 3

    @parameterize_batch_for_data_sources(
        data_source_configs=PANDAS_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_pandas(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = ColumnValuesNonNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, ColumnValuesNonNullCountResult)
        assert metric_result.value == self.NON_NULL_COUNT

    @parameterize_batch_for_data_sources(
        data_source_configs=SPARK_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_spark(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = ColumnValuesNonNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, ColumnValuesNonNullCountResult)
        assert metric_result.value == self.NON_NULL_COUNT

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_sql(self, batch_for_datasource: Batch) -> None:
        batch = batch_for_datasource
        metric = ColumnValuesNonNullCount(column=STRING_COLUMN_NAME)
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, ColumnValuesNonNullCountResult)
        assert metric_result.value == self.NON_NULL_COUNT



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from great_expectations.compatibility.typing_extensions import override
from great_expectations.render import RenderedStringTemplateContent
from great_expectations.render.renderer.content_block.content_block import (
    ContentBlockRenderer,
)
from great_expectations.render.renderer_configuration import RendererConfiguration

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )


class ExpectationStringRenderer(ContentBlockRenderer):
    @classmethod
    @override
    def _missing_content_block_fn(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> List[RenderedStringTemplateContent]:
        renderer_configuration: RendererConfiguration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        return [
            RenderedStringTemplateContent(
                **{  # type: ignore[arg-type] # FIXME CoP
                    "content_block_type": "string_template",
                    "styling": {"parent": {"classes": ["alert", "alert-warning"]}},
                    "string_template": {
                        "template": "$expectation_type(**$kwargs)",
                        "params": {
                            "expectation_type": renderer_configuration.expectation_type,
                            "kwargs": renderer_configuration.kwargs,
                        },
                        "styling": {
                            "params": {
                                "expectation_type": {
                                    "classes": ["badge", "badge-warning"],
                                }
                            }
                        },
                    },
                }
            )
        ]

    @classmethod
    def _diagnostic_status_icon_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        assert result, "Must provide a result object."
        if result.exception_info["raised_exception"]:
            return RenderedStringTemplateContent(
                **{  # type: ignore[arg-type] # FIXME CoP
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "❗"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": [
                                        "fas",
                                        "fa-exclamation-triangle",
                                        "text-warning",
                                    ],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                }
            )

        if result.success:
            return RenderedStringTemplateContent(
                **{  # type: ignore[arg-type] # FIXME CoP
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "✅"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "classes": [
                                        "fas",
                                        "fa-check-circle",
                                        "text-success",
                                    ],
                                    "tag": "i",
                                }
                            }
                        },
                    },
                    "styling": {"parent": {"classes": ["hide-succeeded-validation-target-child"]}},
                }
            )
        else:
            return RenderedStringTemplateContent(
                **{  # type: ignore[arg-type] # FIXME CoP
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$icon",
                        "params": {"icon": "", "markdown_status_icon": "❌"},
                        "styling": {
                            "params": {
                                "icon": {
                                    "tag": "i",
                                    "classes": ["fas", "fa-times", "text-danger"],
                                }
                            }
                        },
                    },
                }
            )



"""
CD to this directory to run these invoke commands.
"""

from __future__ import annotations

import pathlib
from typing import Optional

import invoke
import pandas as pd

TEST_ROOT = pathlib.Path(__file__).parent.parent.parent.resolve(strict=True)


@invoke.task(aliases=["gen-assets"])
def generate_asset_files(
    ctx: invoke.Context,
    source_csv_dir: Optional[str | pathlib.Path] = None,
    limit: int = 12,
    year: str = 2019,
):
    """Generate or re-generate various test asset files."""
    if not source_csv_dir:
        source_csv_dir = TEST_ROOT / "test_sets" / "taxi_yellow_tripdata_samples"

    source_csv_dir = pathlib.Path(source_csv_dir).resolve(strict=True)

    json_asset_dir = pathlib.Path.cwd() / "json_assets"
    json_asset_dir.mkdir(exist_ok=True)
    for i, csv_file in enumerate(source_csv_dir.glob(f"*{year}*.csv"), start=1):
        json_output_file = json_asset_dir / csv_file.with_suffix(".json").name
        print(i, json_output_file.name)
        df: pd.DataFrame = pd.read_csv(csv_file)
        # write in a format that is non-default - default orient is 'index'
        df.to_json(json_output_file, orient="records")
        if i == limit:
            break



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from great_expectations.compatibility import pyspark
from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.metric_function_types import (
    MetricPartialFunctionTypeSuffixes,
)
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.map_metric_provider import (
    MulticolumnMapMetricProvider,
)
from great_expectations.expectations.metrics.map_metric_provider.multicolumn_condition_partial import (  # noqa: E501 # FIXME CoP
    multicolumn_condition_partial,
)
from great_expectations.expectations.metrics.map_metric_provider.multicolumn_function_partial import (  # noqa: E501 # FIXME CoP
    multicolumn_function_partial,
)
from great_expectations.validator.validation_graph import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )


class CompoundColumnsUnique(MulticolumnMapMetricProvider):
    """
    While the support for "PandasExecutionEngine" and "SparkDFExecutionEngine" is accomplished using a compact
    implementation, which combines the "map" and "condition" parts in a single step, the support for
    "SqlAlchemyExecutionEngine" is more detailed.  Thus, the "map" and "condition" parts for "SqlAlchemyExecutionEngine"
    are handled separately, with the "condition" part relying on the "map" part as a metric dependency.
    """  # noqa: E501 # FIXME CoP

    function_metric_name = "compound_columns.count"  # pre-requisite "map" style metric
    condition_metric_name = (
        "compound_columns.unique"  # "condition" style metric required to be implemented by provider
    )
    condition_domain_keys = (
        "batch_id",
        "table",
        "column_list",
        "row_condition",
        "condition_parser",
        "ignore_row_if",
    )

    @multicolumn_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column_list, **kwargs):
        row_wise_cond = ~column_list.duplicated(keep=False)
        return row_wise_cond

    @multicolumn_function_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy_function(self, column_list, **kwargs):
        """
        Computes the "map" between the specified "column_list" (treated as a group so as to model the "compound" aspect)
        and the number of occurrences of every permutation of the values of "column_list" as the grouped subset of all
        rows of the table.  In the present context, the term "compound" refers to having to treat the specified columns
        as unique together (e.g., as a multi-column primary key).  For example, suppose that in the example below, all
        three columns ("A", "B", and "C") of the table are included as part of the "compound" columns list (i.e.,
        column_list = ["A", "B", "C"]):

            A B C _num_rows
            1 1 2 2
            1 2 3 1
            1 1 2 2
            2 2 2 1
            3 2 3 1

        The fourth column, "_num_rows", holds the value of the "map" function -- the number of rows the group occurs in.
        """  # noqa: E501 # FIXME CoP

        # Needed as keys (hence, string valued) to access "ColumnElement" objects contained within the "FROM" clauses.  # noqa: E501 # FIXME CoP
        column_names = kwargs.get("_column_names")

        # Need all columns of the table for the purposes of reporting entire rows satisfying unexpected condition logic.  # noqa: E501 # FIXME CoP
        table_columns = kwargs.get("_table_columns")

        table = kwargs.get(
            "_table"
        )  # Note that here, "table" is of the "sqlalchemy.sql.selectable.Subquery" type.

        # Filipe - 20231114
        # This is a special case that needs to be handled for mysql, where you cannot refer to a temp_table  # noqa: E501 # FIXME CoP
        # more than once in the same query. The solution to this is to perform our operation without the need  # noqa: E501 # FIXME CoP
        # for a sub query. We can do this by using the window function count, to get the number of duplicate  # noqa: E501 # FIXME CoP
        # rows by over partition by the compound unique columns. This will give a table which has the same  # noqa: E501 # FIXME CoP
        # number of rows as the original table, but with an additional column _num_rows column.
        dialect = kwargs.get("_dialect")
        try:
            dialect_name = dialect.dialect.name
        except AttributeError:
            try:
                dialect_name = dialect.name
            except AttributeError:
                dialect_name = ""
        if dialect and dialect_name == "mysql":
            table_columns_selector = [sa.column(column_name) for column_name in table_columns]
            partition_by_columns = (
                sa.func.count()
                .over(partition_by=[sa.column(column) for column in column_names])
                .label("_num_rows")
            )
            count_selector = table_columns_selector + [partition_by_columns]
            original_table_clause = (
                sa.select(*count_selector).select_from(table).alias("original_table_clause")
            )
            return original_table_clause

        # Step-1: Obtain the SQLAlchemy "FromClause" version of the original "table" for the purposes of gaining the  # noqa: E501 # FIXME CoP
        # "FromClause.c" attribute, which is a namespace of all the columns contained within the "FROM" clause (these  # noqa: E501 # FIXME CoP
        # elements are themselves subclasses of the SQLAlchemy "ColumnElement" class).
        table_columns_selector = [sa.column(column_name) for column_name in table_columns]
        original_table_clause = (
            sa.select(*table_columns_selector).select_from(table).alias("original_table_clause")
        )

        # Step-2: "SELECT FROM" the original table, represented by the "FromClause" object, querying all columns of the  # noqa: E501 # FIXME CoP
        # table and the count of occurrences of distinct "compound" (i.e., group, as specified by "column_list") values.  # noqa: E501 # FIXME CoP
        # Give this aggregated group count a distinctive label.
        # Give the resulting sub-query a unique alias in order to disambiguate column names in subsequent queries.  # noqa: E501 # FIXME CoP
        count_selector = column_list + [sa.func.count().label("_num_rows")]
        group_count_query = (
            sa.select(*count_selector)
            .group_by(*column_list)
            .select_from(original_table_clause)
            .alias("group_counts_subquery")
        )

        # The above "group_count_query", if executed, will produce the result set containing the number of rows that  # noqa: E501 # FIXME CoP
        # equals the number of distinct values of the group -- unique grouping (e.g., as in a multi-column primary key).  # noqa: E501 # FIXME CoP
        # Hence, in order for the "_num_rows" column values to provide an entry for each row of the original table, the  # noqa: E501 # FIXME CoP
        # "SELECT FROM" of "group_count_query" must undergo an "INNER JOIN" operation with the "original_table_clause"  # noqa: E501 # FIXME CoP
        # object, whereby all table columns in the two "FromClause" objects must match, respectively, as the conditions.  # noqa: E501 # FIXME CoP
        conditions = sa.and_(
            *(group_count_query.c[name] == original_table_clause.c[name] for name in column_names)
        )
        # noinspection PyProtectedMember
        compound_columns_count_query = (
            sa.select(
                original_table_clause,
                group_count_query.c._num_rows.label("_num_rows"),
            )
            .select_from(
                original_table_clause.join(
                    right=group_count_query, onclause=conditions, isouter=False
                )
            )
            .alias("records_with_grouped_column_counts_subquery")
        )

        # The returned SQLAlchemy "FromClause" "compound_columns_count_query" object realizes the "map" metric function.  # noqa: E501 # FIXME CoP
        return compound_columns_count_query

    @multicolumn_condition_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy_condition(cls, column_list, **kwargs):
        """
        Retrieve the specified "map" metric dependency value as the "FromClause" "compound_columns_count_query" object
        and extract from it -- using the supported SQLAlchemy column access method -- the "_num_rows" columns.  The
        uniqueness of "compound" columns (as a group) is expressed by the "BinaryExpression" "row_wise_cond" returned.

        Importantly, since the "compound_columns_count_query" is the "FromClause" object that incorporates all columns
        of the original table, no additional "FromClause" objects ("select_from") must augment this "condition" metric.
        Other than boolean operations, column access, argument of filtering, and limiting the size of the result set,
        this "row_wise_cond", serving as the main component of the unexpected condition logic, carries along with it
        the entire object hierarchy, making any encapsulating query ready for execution against the database engine.
        """  # noqa: E501 # FIXME CoP

        metrics = kwargs.get("_metrics")
        compound_columns_count_query, _, _ = metrics[
            f"compound_columns.count.{MetricPartialFunctionTypeSuffixes.MAP.value}"
        ]

        # noinspection PyProtectedMember
        row_wise_cond = compound_columns_count_query.c._num_rows < 2  # noqa: PLR2004 # FIXME CoP

        return row_wise_cond

    @multicolumn_condition_partial(engine=SparkDFExecutionEngine)
    def _spark(cls, column_list, **kwargs):
        column_names = column_list.columns
        row_wise_cond = (
            F.count(F.lit(1)).over(pyspark.Window.partitionBy(F.struct(*column_names))) <= 1
        )
        return row_wise_cond

    @classmethod
    @override
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        """
        Returns a dictionary of given metric names and their corresponding configuration, specifying the metric types
        and their respective domains.
        """  # noqa: E501 # FIXME CoP

        dependencies: dict = super()._get_evaluation_dependencies(
            metric=metric,
            configuration=configuration,
            execution_engine=execution_engine,
            runtime_configuration=runtime_configuration,
        )

        if isinstance(execution_engine, SqlAlchemyExecutionEngine):
            if (
                metric.metric_name
                == f"compound_columns.unique.{MetricPartialFunctionTypeSuffixes.CONDITION.value}"
            ):
                dependencies[
                    f"compound_columns.count.{MetricPartialFunctionTypeSuffixes.MAP.value}"
                ] = MetricConfiguration(
                    metric_name=f"compound_columns.count.{MetricPartialFunctionTypeSuffixes.MAP.value}",
                    metric_domain_kwargs=metric.metric_domain_kwargs,
                    metric_value_kwargs=None,
                )

        return dependencies



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Type, Union

from great_expectations.compatibility.pydantic import Field, StrictStr
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.expectations.expectation import (
    BatchExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import COLUMN_DESCRIPTION
from great_expectations.render import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import ordinal, substitute_none_for_missing

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs

EXPECTATION_SHORT_DESCRIPTION = "Checks for the existence of a specified column within a table."
COLUMN_INDEX_DESCRIPTION = (
    "If not None, checks the order of the columns. "
    "The expectation will fail if the column is not in location column_index (zero-indexed)."
)
DATA_QUALITY_ISSUES = [DataQualityIssues.SCHEMA.value]
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]


class ExpectColumnToExist(BatchExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnToExist is a \
    Batch Expectation.

    BatchExpectations are one of the most common types of Expectation. They are evaluated for an entire Batch, and answer a semantic question about the Batch itself.

    Args:
        column (str): {COLUMN_DESCRIPTION}

    Other Parameters:
        column_index (int or None, optional): {COLUMN_INDEX_DESCRIPTION}
        result_format (str or None, optional): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None, optional): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None, optional): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	1.00 	2
            1 	2.30 	5
            2 	4.33 	0

    Passing Case:
        Input:
            ExpectColumnToExist(
                column="test",
                column_index=0
        )

        Output:
            {{
              "exception_info": {{
                "raised_exception": false,
                "exception_traceback": null,
                "exception_message": null
              }},
              "meta": {{}},
              "success": true,
              "result": {{}}
            }}

    Failing Case:
        Input:
            ExpectColumnToExist(
                column="missing_column",
            )

        Output:
            {{
              "exception_info": {{
                "raised_exception": false,
                "exception_traceback": null,
                "exception_message": null
              }},
              "meta": {{}},
              "success": false,
              "result": {{}}
            }}
    """  # noqa: E501 # FIXME CoP

    column: StrictStr = Field(min_length=1, description=COLUMN_DESCRIPTION)
    column_index: Union[int, SuiteParameterDict, None] = Field(
        default=None, description=COLUMN_INDEX_DESCRIPTION
    )

    # This dictionary contains metadata for display in the public gallery
    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "table expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    metric_dependencies = ("table.columns",)
    success_keys = (
        "column",
        "column_index",
    )
    domain_keys = (
        "batch_id",
        "table",
    )
    args_keys = ("column", "column_index")

    class Config:
        title = "Expect column to exist"

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type[ExpectColumnToExist]) -> None:
            BatchExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    @override
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("column_index", RendererValueType.NUMBER),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if not params.column_index:
            if renderer_configuration.include_column_name:
                template_str = "$column is a required field."
            else:
                template_str = "is a required field."
        else:
            renderer_configuration.add_param(
                name="column_indexth",
                param_type=RendererValueType.STRING,
                value=ordinal(params.column_index.value),
            )
            if renderer_configuration.include_column_name:
                template_str = "$column must be the $column_indexth field."
            else:
                template_str = "must be the $column_indexth field."

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @override
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> list[RenderedStringTemplateContent]:
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,  # type: ignore[union-attr] # FIXME: could be None
            ["column", "column_index"],
        )

        if params["column_index"] is None:
            if include_column_name:
                template_str = "$column is a required field."
            else:
                template_str = "is a required field."
        else:
            params["column_indexth"] = ordinal(params["column_index"])
            if include_column_name:
                template_str = "$column must be the $column_indexth field."
            else:
                template_str = "must be the $column_indexth field."

        return [
            RenderedStringTemplateContent(
                content_block_type="string_template",
                string_template={
                    "template": template_str,
                    "params": params,
                    "styling": styling,
                },
            )
        ]

    @override
    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        actual_columns = metrics["table.columns"]
        expected_column_name = self._get_success_kwargs().get("column")
        expected_column_index = self._get_success_kwargs().get("column_index")

        if expected_column_index:
            try:
                success = actual_columns[expected_column_index] == expected_column_name
            except IndexError:
                success = False
        else:
            success = expected_column_name in actual_columns

        return {"success": success}



from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING

import boto3
import botocore
import pytest

from great_expectations.datasource.fluent import PandasDBFSDatasource
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import CSVAsset
from great_expectations.datasource.fluent.dynamic_pandas import PANDAS_VERSION
from tests.test_utils import create_files_in_directory

if TYPE_CHECKING:
    from pyfakefs.fake_filesystem import FakeFilesystem

    from great_expectations.data_context import FileDataContext


logger = logging.getLogger(__file__)


# apply markers to entire test module
pytestmark = [
    pytest.mark.skipif(
        PANDAS_VERSION < 1.2, reason=f"Fluent pandas not supported on {PANDAS_VERSION}"
    )
]


@pytest.fixture
def pandas_dbfs_datasource(
    empty_data_context: FileDataContext, fs: FakeFilesystem
) -> PandasDBFSDatasource:
    # Copy boto modules into fake filesystem (see https://github.com/spulec/moto/issues/1682#issuecomment-645016188)
    for module in [boto3, botocore]:
        module_dir = pathlib.Path(module.__file__).parent
        fs.add_real_directory(module_dir, lazy_read=False)

    base_directory: str = "/dbfs/great_expectations"
    fs.create_dir(base_directory)

    fs.create_dir(empty_data_context.root_directory)

    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "yellow_tripdata_sample_2024-01.csv",
            "yellow_tripdata_sample_2024-02.csv",
            "yellow_tripdata_sample_2024-03.csv",
            "yellow_tripdata_sample_2024-04.csv",
            "yellow_tripdata_sample_2024-05.csv",
            "yellow_tripdata_sample_2024-06.csv",
            "yellow_tripdata_sample_2024-07.csv",
            "yellow_tripdata_sample_2024-08.csv",
            "yellow_tripdata_sample_2024-09.csv",
            "yellow_tripdata_sample_2024-10.csv",
            "yellow_tripdata_sample_2024-11.csv",
            "yellow_tripdata_sample_2024-12.csv",
        ],
    )

    pandas_dbfs_datasource = PandasDBFSDatasource(
        name="pandas_dbfs_datasource",
        base_directory=pathlib.Path(base_directory),
    )
    pandas_dbfs_datasource._data_context = empty_data_context

    return pandas_dbfs_datasource


@pytest.mark.filesystem
def test_construct_pandas_dbfs_datasource(pandas_dbfs_datasource: PandasDBFSDatasource):
    assert pandas_dbfs_datasource.name == "pandas_dbfs_datasource"


@pytest.mark.filesystem
def test_add_csv_asset_to_datasource(pandas_dbfs_datasource: PandasDBFSDatasource):
    asset = pandas_dbfs_datasource.add_csv_asset(
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.filesystem
def test_construct_csv_asset_directly():
    # noinspection PyTypeChecker
    asset = CSVAsset(
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.filesystem
def test_get_batch_list_from_fully_specified_batch_request(
    pandas_dbfs_datasource: PandasDBFSDatasource,
):
    asset = pandas_dbfs_datasource.add_csv_asset(
        name="csv_asset",
    )
    regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    batch_def = asset.add_batch_definition_monthly(name="batch def", regex=regex)
    batch_parameters = {"year": "2024", "month": "05"}
    batch = batch_def.get_batch(batch_parameters=batch_parameters)

    assert batch.batch_request.datasource_name == pandas_dbfs_datasource.name
    assert batch.batch_request.data_asset_name == asset.name
    path = "yellow_tripdata_sample_2024-05.csv"
    assert batch.batch_request.options == {"path": path, "year": "2024", "month": "05"}
    assert batch.metadata == {"path": path, "year": "2024", "month": "05"}
    assert batch.id == "pandas_dbfs_datasource-csv_asset-year_2024-month_05"



from typing import Dict, List

import pytest

from great_expectations.data_context import CloudDataContext
from great_expectations.datasource.fluent import BatchRequest
from great_expectations.datasource.fluent.interfaces import Batch, DataAsset
from great_expectations.experimental.metric_repository.metric_list_metric_retriever import (
    MetricListMetricRetriever,
)
from great_expectations.experimental.metric_repository.metrics import (
    ColumnMetric,
    MetricException,
    MetricTypes,
    TableMetric,
)
from great_expectations.validator.exception_info import ExceptionInfo
from great_expectations.validator.validator import Validator

pytestmark = pytest.mark.unit

import logging

from pytest_mock import MockerFixture

LOGGER = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def mock_data_asset(mocker):
    data_asset = mocker.Mock(spec=DataAsset)
    data_asset.name = "some_data_asset_name"
    return data_asset


@pytest.fixture(scope="function")
def mock_validator(mocker, mock_batch):
    validator = mocker.Mock(spec=Validator)
    validator.active_batch = mock_batch
    return validator


@pytest.fixture(scope="function")
def mock_context(mocker, mock_validator):
    context = mocker.Mock(spec=CloudDataContext)
    context.get_validator.return_value = mock_validator
    return context


@pytest.fixture(scope="function")
def mock_batch(mocker, mock_data_asset):
    batch = mocker.Mock(spec=Batch)
    batch.id = "batch_id"
    batch.data_asset = mock_data_asset
    return batch


@pytest.fixture(scope="function")
def metric_retriever(mock_context):
    return MetricListMetricRetriever(context=mock_context)


@pytest.fixture(scope="function")
def mock_batch_request(mocker):
    batch_request = mocker.Mock(spec=BatchRequest)
    batch_request.data_asset_name = "some_data_asset_name"
    return batch_request


@pytest.fixture(scope="function")
def mock_batch_request_variant(mocker):
    batch_request = mocker.Mock(spec=BatchRequest)
    batch_request.data_asset_name = "other_data_asset_name"
    return batch_request


def test_get_metrics_table_metrics_only(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    computed_metrics = {
        ("table.row_count", (), ()): 2,
        ("table.columns", (), ()): ["col1", "col2"],
        ("table.column_types", (), ()): [
            {"name": "col1", "type": "float"},
            {"name": "col2", "type": "float"},
        ],
    }
    table_metrics_list = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
    ]
    aborted_metrics: Dict[str, str] = {}
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )

    metrics = metric_retriever.get_metrics(
        batch_request=mock_batch_request,
        metric_list=table_metrics_list,
    )
    assert metrics == [
        TableMetric[int](
            batch_id="batch_id",
            metric_name="table.row_count",
            value=2,
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.columns",
            value=["col1", "col2"],
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.column_types",
            value=[
                {"name": "col1", "type": "float"},
                {"name": "col2", "type": "float"},
            ],
            exception=None,
        ),
    ]


def test_get_metrics_full_list(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    computed_metrics = {
        ("table.row_count", (), ()): 2,
        ("table.columns", (), ()): ["col1", "col2"],
        ("table.column_types", (), ()): [
            {"name": "col1", "type": "float"},
            {"name": "col2", "type": "float"},
        ],
        ("column.min", "column=col1", ()): 2.5,
        ("column.min", "column=col2", ()): 2.7,
        ("column.max", "column=col1", ()): 5.5,
        ("column.max", "column=col2", ()): 5.7,
        ("column.mean", "column=col1", ()): 2.5,
        ("column.mean", "column=col2", ()): 2.7,
        ("column.median", "column=col1", ()): 2.5,
        ("column.median", "column=col2", ()): 2.7,
        ("column_values.null.count", "column=col1", ()): 1,
        ("column_values.null.count", "column=col2", ()): 1,
    }
    metrics_list = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
        MetricTypes.COLUMN_MAX,
        MetricTypes.COLUMN_MEAN,
        MetricTypes.COLUMN_MEDIAN,
        MetricTypes.COLUMN_NULL_COUNT,
    ]
    aborted_metrics: Dict[str, str] = {}
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )

    patch_get_numeric_column_names_with(mocker, ["col1", "col2"])
    patch_get_timestamp_column_names_with(mocker, [])
    metrics = metric_retriever.get_metrics(
        batch_request=mock_batch_request,
        metric_list=metrics_list,
    )

    assert metrics == [
        TableMetric[int](
            batch_id="batch_id", metric_name="table.row_count", value=2, exception=None
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.columns",
            value=["col1", "col2"],
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.column_types",
            value=[
                {"name": "col1", "type": "float"},
                {"name": "col2", "type": "float"},
            ],
            exception=None,
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.max",
            value=5.5,
            exception=None,
            column="col1",
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.max",
            value=5.7,
            exception=None,
            column="col2",
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.mean",
            value=2.5,
            exception=None,
            column="col1",
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.mean",
            value=2.7,
            exception=None,
            column="col2",
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.median",
            value=2.5,
            exception=None,
            column="col1",
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.median",
            value=2.7,
            exception=None,
            column="col2",
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.min",
            value=2.5,
            exception=None,
            column="col1",
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.min",
            value=2.7,
            exception=None,
            column="col2",
        ),
        ColumnMetric[int](
            batch_id="batch_id",
            metric_name="column_values.null.count",
            value=1,
            exception=None,
            column="col1",
        ),
        ColumnMetric[int](
            batch_id="batch_id",
            metric_name="column_values.null.count",
            value=1,
            exception=None,
            column="col2",
        ),
    ]


def test_column_metrics_not_returned_if_column_types_missing(
    mocker: MockerFixture,
    caplog,
    mock_context,
    mock_validator,
    mock_batch_request,
    metric_retriever,
):
    computed_metrics = {
        ("table.row_count", (), ()): 2,
        ("table.columns", (), ()): ["timestamp_col"],
    }
    metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        # MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
        MetricTypes.COLUMN_MAX,
        MetricTypes.COLUMN_NULL_COUNT,
    ]
    aborted_metrics = {}
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )

    patch_get_numeric_column_names_with(mocker, [])
    patch_get_timestamp_column_names_with(mocker, ["timestamp_col"])
    metrics = metric_retriever.get_metrics(
        batch_request=mock_batch_request, metric_list=metrics_list
    )

    assert metrics == [
        TableMetric[int](
            batch_id="batch_id",
            metric_name="table.row_count",
            value=2,
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.columns",
            value=["timestamp_col"],
            exception=None,
        ),
    ]
    assert "TABLE_COLUMN_TYPES metric is required to compute column metrics." in caplog.text


def patch_get_numeric_column_names_with(mocker, return_value):
    mocker.patch(
        f"{MetricListMetricRetriever.__module__}.{MetricListMetricRetriever.__name__}._get_numeric_column_names",
        return_value=return_value,
    )


def test_get_metrics_metrics_missing(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    """This test is meant to simulate metrics missing from the computed metrics."""
    mock_computed_metrics = {
        # ("table.row_count", (), ()): 2, # Missing table.row_count metric
        ("table.columns", (), ()): ["col1", "col2"],
        ("table.column_types", (), ()): [
            {"name": "col1", "type": "float"},
            {"name": "col2", "type": "float"},
        ],
        # ("column.min", "column=col1", ()): 2.5, # Missing column.min metric for col1
        ("column.min", "column=col2", ()): 2.7,
    }

    metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
    ]
    mock_aborted_metrics = {}
    mock_validator.compute_metrics.return_value = (
        mock_computed_metrics,
        mock_aborted_metrics,
    )
    patch_get_numeric_column_names_with(mocker, ["col1", "col2"])
    patch_get_timestamp_column_names_with(mocker, [])
    metrics = metric_retriever.get_metrics(
        batch_request=mock_batch_request, metric_list=metrics_list
    )
    assert metrics == [
        TableMetric[int](
            batch_id="batch_id",
            metric_name="table.row_count",
            value=None,
            exception=MetricException(
                type="Not found",
                message="Metric was not successfully computed but exception was not found.",
            ),
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.columns",
            value=["col1", "col2"],
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.column_types",
            value=[
                {"name": "col1", "type": "float"},
                {"name": "col2", "type": "float"},
            ],
            exception=None,
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.min",
            value=None,
            exception=MetricException(
                type="Not found",
                message="Metric was not successfully computed but exception was not found.",
            ),
            column="col1",
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.min",
            value=2.7,
            exception=None,
            column="col2",
        ),
    ]


def patch_get_timestamp_column_names_with(mocker, return_value):
    mocker.patch(
        f"{MetricListMetricRetriever.__module__}.{MetricListMetricRetriever.__name__}._get_timestamp_column_names",
        return_value=return_value,
    )


def test_get_metrics_with_exception(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    """This test is meant to simulate failed metrics in the computed metrics."""

    exception_info = ExceptionInfo(
        exception_traceback="test exception traceback",
        exception_message="test exception message",
        raised_exception=True,
    )
    aborted_metrics = {
        ("table.row_count", (), ()): {
            "metric_configuration": {},  # Leaving out for brevity
            "num_failures": 3,
            "exception_info": exception_info,
        },
        ("column.min", "column=col1", ()): {
            "metric_configuration": {},  # Leaving out for brevity
            "num_failures": 3,
            "exception_info": exception_info,
        },
    }
    computed_metrics = {
        # ("table.row_count", (), ()): 2, # Error in table.row_count metric
        ("table.columns", (), ()): ["col1", "col2"],
        ("table.column_types", (), ()): [
            {"name": "col1", "type": "float"},
            {"name": "col2", "type": "float"},
        ],
    }
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )
    metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
    ]

    metrics = metric_retriever.get_metrics(
        batch_request=mock_batch_request, metric_list=metrics_list
    )

    assert metrics == [
        TableMetric[int](
            batch_id="batch_id",
            metric_name="table.row_count",
            value=None,
            exception=MetricException(type="Unknown", message="test exception message"),
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.columns",
            value=["col1", "col2"],
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.column_types",
            value=[
                {"name": "col1", "type": "float"},
                {"name": "col2", "type": "float"},
            ],
            exception=None,
        ),
    ]


def test_get_metrics_with_column_type_missing(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    """This test is meant to simulate failed metrics in the computed metrics."""

    exception_info = ExceptionInfo(
        exception_traceback="test exception traceback",
        exception_message="test exception message",
        raised_exception=True,
    )

    aborted_metrics = {
        ("table.row_count", (), ()): {
            "metric_configuration": {},  # Leaving out for brevity
            "num_failures": 3,
            "exception_info": exception_info,
        },
        ("column.min", "column=col1", ()): {
            "metric_configuration": {},  # Leaving out for brevity
            "num_failures": 3,
            "exception_info": exception_info,
        },
    }

    computed_metrics = {
        # ("table.row_count", (), ()): 2, # Error in table.row_count metric
        ("table.columns", (), ()): ["col1", "col2"],
        ("table.column_types", (), ()): [
            {"name": "col1", "type": "float"},
            {
                "name": "col2",
            },  # Missing type for col2
        ],
        # ("column.min", "column=col1", ()): 2.5, # Error in column.min metric for col1
        ("column.min", "column=col2", ()): 2.7,
    }
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )
    metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
    ]

    patch_get_numeric_column_names_with(mocker, ["col1", "col2"])
    patch_get_timestamp_column_names_with(mocker, [])

    metrics = metric_retriever.get_metrics(
        batch_request=mock_batch_request, metric_list=metrics_list
    )
    assert metrics == [
        TableMetric[int](
            batch_id="batch_id",
            metric_name="table.row_count",
            value=None,
            exception=MetricException(type="Unknown", message="test exception message"),
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.columns",
            value=["col1", "col2"],
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.column_types",
            value=[
                {"name": "col1", "type": "float"},
                {
                    "name": "col2",
                },  # Note: No type for col2
            ],
            exception=None,
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.min",
            column="col1",
            value=None,
            exception=MetricException(type="Unknown", message="test exception message"),
        ),
        ColumnMetric[float](
            batch_id="batch_id",
            metric_name="column.min",
            column="col2",
            value=2.7,
            exception=None,
        ),
    ]


def test_get_metrics_with_timestamp_columns(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    computed_metrics = {
        ("table.row_count", (), ()): 2,
        ("table.columns", (), ()): ["timestamp_col"],
        ("table.column_types", (), ()): [
            {"name": "timestamp_col", "type": "TIMESTAMP_NTZ"},
        ],
        ("column.min", "column=timestamp_col", ()): "2023-01-01T00:00:00",
        ("column.max", "column=timestamp_col", ()): "2023-12-31T00:00:00",
        ("column_values.null.count", "column=timestamp_col", ()): 1,
    }
    metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
        MetricTypes.COLUMN_MAX,
        MetricTypes.COLUMN_NULL_COUNT,
    ]
    aborted_metrics = {}
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )
    patch_get_numeric_column_names_with(mocker, [])
    patch_get_timestamp_column_names_with(mocker, ["timestamp_col"])

    metrics = metric_retriever.get_metrics(
        batch_request=mock_batch_request, metric_list=metrics_list
    )

    assert metrics == [
        TableMetric[int](
            batch_id="batch_id",
            metric_name="table.row_count",
            value=2,
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.columns",
            value=["timestamp_col"],
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.column_types",
            value=[{"name": "timestamp_col", "type": "TIMESTAMP_NTZ"}],
            exception=None,
        ),
        ColumnMetric[str](
            batch_id="batch_id",
            metric_name="column.max",
            value="2023-12-31T00:00:00",
            exception=None,
            column="timestamp_col",
        ),
        ColumnMetric[str](
            batch_id="batch_id",
            metric_name="column.min",
            value="2023-01-01T00:00:00",
            exception=None,
            column="timestamp_col",
        ),
        ColumnMetric[int](
            batch_id="batch_id",
            metric_name="column_values.null.count",
            value=1,
            exception=None,
            column="timestamp_col",
        ),
    ]


def test_get_metrics_only_gets_a_validator_once(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    aborted_metrics = {}

    computed_metrics = {
        ("table.row_count", (), ()): 2,
        ("table.columns", (), ()): ["col1", "col2"],
        ("table.column_types", (), ()): [
            {"name": "col1", "type": "float"},
            {"name": "col2", "type": "float"},
        ],
    }
    metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
    ]
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )
    # No column filtering needed for table-only metrics
    metric_retriever.get_metrics(batch_request=mock_batch_request, metric_list=metrics_list)

    mock_context.get_validator.assert_called_once_with(batch_request=mock_batch_request)


def test_get_metrics_only_gets_new_validator_on_asset_change(
    mocker: MockerFixture,
    mock_context,
    mock_validator,
    mock_batch_request_variant,
    metric_retriever,
):
    aborted_metrics = {}

    computed_metrics = {
        ("table.row_count", (), ()): 2,
        ("table.columns", (), ()): ["col1", "col2"],
        ("table.column_types", (), ()): [
            {"name": "col1", "type": "float"},
            {"name": "col2", "type": "float"},
        ],
    }
    metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
    ]
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )
    # No column filtering needed for table-only metrics
    metric_retriever.get_metrics(batch_request=mock_batch_request_variant, metric_list=metrics_list)

    assert mock_context.get_validator.call_count == 4
    mock_context.get_validator.assert_called_with(batch_request=mock_batch_request_variant)


def test_get_metrics_with_no_metrics(
    mock_context, mock_validator, mock_batch_request, metric_retriever
):
    computed_metrics = {}
    metrics_list: List[MetricTypes] = []
    aborted_metrics = {}
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )
    with pytest.raises(ValueError):
        metric_retriever.get_metrics(batch_request=mock_batch_request, metric_list=metrics_list)


def test_valid_metric_types_true(mock_context, metric_retriever):
    valid_metric_types = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
        MetricTypes.COLUMN_MAX,
        MetricTypes.COLUMN_MEAN,
        MetricTypes.COLUMN_MEDIAN,
        MetricTypes.COLUMN_NULL_COUNT,
    ]
    assert metric_retriever._check_valid_metric_types(valid_metric_types) is True


def test_valid_metric_types_false(mocker: MockerFixture, mock_context, metric_retriever):
    invalid_metric_type = ["I_am_invalid"]
    assert metric_retriever._check_valid_metric_types(invalid_metric_type) is False


def test_column_metrics_in_metrics_list_only_table_metrics(
    mocker: MockerFixture, mock_context, metric_retriever
):
    table_metrics_only = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
    ]
    assert metric_retriever._column_metrics_in_metric_list(table_metrics_only) is False


def test_column_metrics_in_metrics_list_with_column_metrics(
    mocker: MockerFixture, mock_context, metric_retriever
):
    metrics_list_with_column_metrics = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
    ]
    assert metric_retriever._column_metrics_in_metric_list(metrics_list_with_column_metrics) is True


def test_get_table_column_types(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    computed_metrics = {
        ("table.column_types", (), ()): [
            {"name": "col1", "type": "float"},
            {"name": "col2", "type": "float"},
        ],
    }
    aborted_metrics = {}
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )
    ret = metric_retriever._get_table_column_types(mock_batch_request)
    print(ret)


def test_get_table_columns(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    computed_metrics = {
        ("table.columns", (), ()): ["col1", "col2"],
    }
    aborted_metrics = {}
    mock_validator.compute_metrics.return_value = (computed_metrics, aborted_metrics)

    ret = metric_retriever._get_table_columns(mock_batch_request)
    assert ret == TableMetric[List[str]](
        batch_id="batch_id",
        metric_name="table.columns",
        value=["col1", "col2"],
        exception=None,
    )


def test_get_table_row_count(
    mocker: MockerFixture, mock_context, mock_validator, mock_batch_request, metric_retriever
):
    computed_metrics = {("table.row_count", (), ()): 2}
    aborted_metrics = {}
    mock_validator.compute_metrics.return_value = (computed_metrics, aborted_metrics)

    ret = metric_retriever._get_table_row_count(mock_batch_request)
    assert ret == TableMetric[int](
        batch_id="batch_id",
        metric_name="table.row_count",
        value=2,
        exception=None,
    )


def test_get_metrics_with_timestamp_columns_exclude_time(
    mocker: MockerFixture, mock_context, mock_validator, metric_retriever, mock_batch_request
):
    computed_metrics = {
        ("table.row_count", (), ()): 2,
        ("table.columns", (), ()): ["timestamp_col", "time_col"],
        ("table.column_types", (), ()): [
            {"name": "timestamp_col", "type": "TIMESTAMP_NTZ"},
            {"name": "time_col", "type": "TIME"},
        ],
        ("column.min", "column=timestamp_col", ()): "2023-01-01T00:00:00",
        ("column.max", "column=timestamp_col", ()): "2023-12-31T00:00:00",
        ("column_values.null.count", "column=timestamp_col", ()): 1,
        ("column_values.null.count", "column=time_col", ()): 1,
    }
    metrics_list: List[MetricTypes] = [
        MetricTypes.TABLE_ROW_COUNT,
        MetricTypes.TABLE_COLUMNS,
        MetricTypes.TABLE_COLUMN_TYPES,
        MetricTypes.COLUMN_MIN,
        MetricTypes.COLUMN_MAX,
        MetricTypes.COLUMN_NULL_COUNT,
    ]
    aborted_metrics = {}
    mock_validator.compute_metrics.return_value = (
        computed_metrics,
        aborted_metrics,
    )
    patch_get_numeric_column_names_with(mocker, [])
    patch_get_timestamp_column_names_with(mocker, ["timestamp_col"])
    metrics = metric_retriever.get_metrics(
        batch_request=mock_batch_request, metric_list=metrics_list
    )

    assert metrics == [
        TableMetric[int](
            batch_id="batch_id",
            metric_name="table.row_count",
            value=2,
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.columns",
            value=["timestamp_col", "time_col"],
            exception=None,
        ),
        TableMetric[List[str]](
            batch_id="batch_id",
            metric_name="table.column_types",
            value=[
                {"name": "timestamp_col", "type": "TIMESTAMP_NTZ"},
                {"name": "time_col", "type": "TIME"},
            ],
            exception=None,
        ),
        ColumnMetric[str](
            batch_id="batch_id",
            metric_name="column.max",
            value="2023-12-31T00:00:00",
            exception=None,
            column="timestamp_col",
        ),
        ColumnMetric[str](
            batch_id="batch_id",
            metric_name="column.min",
            value="2023-01-01T00:00:00",
            exception=None,
            column="timestamp_col",
        ),
        ColumnMetric[int](
            batch_id="batch_id",
            metric_name="column_values.null.count",
            value=1,
            exception=None,
            column="timestamp_col",
        ),
        ColumnMetric[int](
            batch_id="batch_id",
            metric_name="column_values.null.count",
            value=1,
            exception=None,
            column="time_col",
        ),
    ]



import logging
from typing import Any

import pytest

import great_expectations as gx
from great_expectations.compatibility.pyspark import ConnectDataFrame, Row, SparkConnectSession
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.exceptions.exceptions import BuildBatchRequestError

logger = logging.getLogger(__name__)


pytestmark = pytest.mark.spark_connect

DATAFRAME_VALUES = [1, 2, 3]


@pytest.fixture
def spark_validation_definition(
    ephemeral_context_with_defaults: AbstractDataContext,
) -> ValidationDefinition:
    context = ephemeral_context_with_defaults
    bd = (
        context.data_sources.add_spark(name="spark-connect-ds")
        .add_dataframe_asset(name="spark-connect-asset")
        .add_batch_definition_whole_dataframe(name="spark-connect-bd")
    )
    suite = context.suites.add(
        gx.ExpectationSuite(
            name="spark-connect-suite",
            expectations=[
                gx.expectations.ExpectColumnValuesToBeInSet(
                    column="column",
                    value_set=DATAFRAME_VALUES,
                ),
            ],
        )
    )
    return context.validation_definitions.add(
        gx.ValidationDefinition(name="spark-connect-vd", suite=suite, data=bd)
    )


def test_spark_connect(
    spark_connect_session: SparkConnectSession,
    spark_validation_definition: ValidationDefinition,
):
    df = spark_connect_session.createDataFrame(
        [Row(column=x) for x in DATAFRAME_VALUES],
    )
    assert isinstance(df, ConnectDataFrame)

    results = spark_validation_definition.run(batch_parameters={"dataframe": df})

    assert results.success


@pytest.mark.parametrize("not_a_dataframe", [None, 1, "string", 1.0, True])
def test_error_messages_if_we_get_an_invalid_dataframe(
    not_a_dataframe: Any,
    spark_validation_definition: ValidationDefinition,
):
    with pytest.raises(
        BuildBatchRequestError, match="Cannot build batch request without a Spark DataFrame."
    ):
        spark_validation_definition.run(batch_parameters={"dataframe": not_a_dataframe})


def test_spark_connect_with_spark_connect_session_factory_method(
    spark_validation_definition: ValidationDefinition,
):
    """This test demonstrates that SparkConnectionSession can be used to create a session.

    This test is being added because in some scenarios, this appeared to fail, but it was
    the result of other active spark sessions.
    """
    spark_connect_session = SparkConnectSession.builder.remote("sc://localhost:15002").getOrCreate()
    assert isinstance(spark_connect_session, SparkConnectSession)
    df = spark_connect_session.createDataFrame(
        [Row(column=x) for x in DATAFRAME_VALUES],
    )

    results = spark_validation_definition.run(batch_parameters={"dataframe": df})

    assert results.success



from typing import Any

from great_expectations.compatibility.pydantic import BaseModel
from great_expectations.metrics.batch.batch import BatchMetric
from great_expectations.metrics.metric_results import MetricResult


class ColumnType(BaseModel):
    name: str
    type: Any


class BatchColumnTypesResult(MetricResult[list[ColumnType]]): ...


class BatchColumnTypes(BatchMetric[BatchColumnTypesResult]):
    """Table schema"""

    name = "table.column_types"



import random
import string

import pandas as pd
import pytest

from great_expectations.core import (
    ExpectationValidationResult,
)
from great_expectations.data_context import CloudDataContext
from great_expectations.render import RenderedAtomicContent
from great_expectations.validator.validator import Validator


@pytest.mark.xfail(
    reason="add_or_update not responsible for rendered content - rewrite test for new suites factory"  # noqa: E501 # FIXME CoP
)
@pytest.mark.cloud
def test_cloud_backed_data_context_expectation_validation_result_include_rendered_content(
    empty_cloud_context_fluent: CloudDataContext,
) -> None:
    """
    All CloudDataContexts should save an ExpectationValidationResult with rendered_content by default.
    """  # noqa: E501 # FIXME CoP
    context = empty_cloud_context_fluent

    df = pd.DataFrame([1, 2, 3, 4, 5])
    suite_name = f"test_suite_{''.join(random.choice(string.ascii_letters + string.digits) for _ in range(8))}"  # noqa: E501 # FIXME CoP

    data_asset = context.data_sources.pandas_default.add_dataframe_asset(
        name="my_dataframe_asset",
    )
    validator: Validator = context.get_validator(
        batch_request=data_asset.build_batch_request(options={"dataframe": df}),
        create_expectation_suite_with_name=suite_name,
    )

    expectation_validation_result: ExpectationValidationResult = (
        validator.expect_table_row_count_to_equal(value=10)
    )

    for rendered_content in expectation_validation_result.rendered_content:
        assert isinstance(rendered_content, RenderedAtomicContent)

    for rendered_content in expectation_validation_result.expectation_config.rendered_content:
        assert isinstance(rendered_content, RenderedAtomicContent)



from typing import Any, Dict, List

import numpy as np
import pandas as pd
import pytest

from great_expectations.compatibility.sqlalchemy_compatibility_wrappers import (
    add_dataframe_to_db,
)
from great_expectations.execution_engine import (
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.execution_engine.sqlalchemy_batch_data import (
    SqlAlchemyBatchData,
)
from great_expectations.expectations.metrics import (
    MapMetricProvider,
)
from great_expectations.expectations.metrics.map_metric_provider.column_map_condition_auxilliary_methods import (  # noqa: E501 # FIXME CoP
    _spark_column_map_condition_values,
    _sqlalchemy_column_map_condition_values,
)
from great_expectations.validator.metric_configuration import MetricConfiguration
from tests.expectations.test_util import get_table_columns_metric


@pytest.fixture
def mini_taxi_df() -> pd.DataFrame:
    """
    Returns: pandas dataframe that contains a small selection of columns and rows from taxi_data, for unittesting.
    """  # noqa: E501 # FIXME CoP
    df = pd.DataFrame(
        {
            "pk_1": [0, 1, 2, 3, 4],
            "vendor_id": [1, 1, 1, 1, 1],
            "pickup_datetime": [
                "2019-01-15 3:36:12",
                "2019-01-25 18:20:32",
                "2019-01-05 6:47:31",
                "2019-01-09 15:08:02",
                "2019-01-25 18:49:51",
            ],
            "dropoff_datetime": [
                "2019-01-15 3:42:19",
                "2019-01-25 18:26:55",
                "2019-01-05 6:52:19",
                "2019-01-09 15:20:17",
                "2019-01-25 18:56:44",
            ],
            "trip_distance": [1, 0.8, 1.1, 2.5, 0.8],
            "tip_amount": [1.95, 1.55, 0, 3, 1.65],
            "total_amount": [9.75, 9.35, 6.8, 14.8, 9.95],
        }
    )
    return df


@pytest.fixture
def sql_execution_engine_with_mini_taxi_loaded(sa, mini_taxi_df):
    sqlite_engine = sa.create_engine("sqlite://")
    dataframe = mini_taxi_df
    add_dataframe_to_db(
        df=dataframe,
        name="test_table",
        con=sqlite_engine,
        index=False,
    )
    execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(engine=sqlite_engine)
    return execution_engine


@pytest.fixture
def spark_execution_engine_with_mini_taxi_loaded(spark_session, mini_taxi_df):
    conf: List[tuple] = spark_session.sparkContext.getConf().getAll()
    spark_config: Dict[str, Any] = dict(conf)

    pandas_df = mini_taxi_df
    spark_df = spark_session.createDataFrame(
        [
            tuple(
                None if isinstance(x, (float, int)) and np.isnan(x) else x for x in record.tolist()
            )
            for record in pandas_df.to_records(index=False)
        ],
        pandas_df.columns.tolist(),
    )
    execution_engine: SparkDFExecutionEngine = SparkDFExecutionEngine(
        spark_config=spark_config,
    )
    execution_engine.load_batch_data(batch_id="1234", batch_data=spark_df)
    return execution_engine


@pytest.fixture
def sql_execution_engine_with_mini_taxi_table_name(sa, sql_execution_engine_with_mini_taxi_loaded):
    execution_engine = sql_execution_engine_with_mini_taxi_loaded
    # BatchData created with `table_name`
    batch_data = SqlAlchemyBatchData(
        execution_engine=execution_engine,
        table_name="test_table",
    )
    execution_engine.load_batch_data("__", batch_data)
    return execution_engine


@pytest.fixture
def sql_execution_engine_with_mini_taxi_query(sa, sql_execution_engine_with_mini_taxi_loaded):
    execution_engine = sql_execution_engine_with_mini_taxi_loaded
    # BatchData created with query
    batch_data = SqlAlchemyBatchData(
        execution_engine=execution_engine,
        query="SELECT * FROM test_table",
    )
    execution_engine.load_batch_data("__", batch_data)
    return execution_engine


@pytest.fixture
def sql_execution_engine_with_mini_taxi_selectable(sa, sql_execution_engine_with_mini_taxi_loaded):
    execution_engine = sql_execution_engine_with_mini_taxi_loaded
    # BatchData created with Selectable
    batch_data = SqlAlchemyBatchData(
        execution_engine=execution_engine,
        selectable=sa.select(sa.text("*")).select_from(sa.table("test_table")),
    )
    execution_engine.load_batch_data("__", batch_data)
    return execution_engine


# We calculate the column_values.between.condition with min value 0.0 and max value 10.0.
# when row_condition is col("pk_1")!=0 _sqlalchemy_column_map_condition_values() method will return [14.8] because it will run against all rows of mini_taxi_df and find the total_amount values that out of range (0 < x < 10.0).  # noqa: E501 # FIXME CoP
# when row_condition is col("pk_1")==0 _sqlalchemy_column_map_condition_values() method will return [] because it will only run against a single row of mini_taxi_df (where pk_1==0), and that total_amount value is within our range (9.75).  # noqa: E501 # FIXME CoP
@pytest.mark.sqlite
@pytest.mark.parametrize(
    "execution_engine_fixture_name",
    (
        "sql_execution_engine_with_mini_taxi_table_name",
        "sql_execution_engine_with_mini_taxi_query",
        "sql_execution_engine_with_mini_taxi_selectable",
    ),
)
@pytest.mark.parametrize(
    "metric_domain_kwargs, expected_result",
    [
        (
            {
                "column": "total_amount",
                "row_condition": 'col("pk_1")==0',
                "condition_parser": "great_expectations",
            },
            [],
        ),
        (
            {
                "column": "total_amount",
                "row_condition": 'col("pk_1")!=0',
                "condition_parser": "great_expectations",
            },
            [14.8],
        ),
    ],
)
def test_sqlalchemy_column_map_condition_values(
    execution_engine_fixture_name, metric_domain_kwargs, expected_result, request, sa
):
    execution_engine = request.getfixturevalue(execution_engine_fixture_name)
    metric_value_kwargs = {
        "min_value": 0,
        "max_value": 10.0,
        "strict_min": False,
        "strict_max": False,
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        },
    }

    desired_metric = MetricConfiguration(
        metric_name="column_values.between.condition",
        metric_domain_kwargs=metric_domain_kwargs,
        metric_value_kwargs=metric_value_kwargs,
    )

    # table.columns metric has to be calculated and loaded first, because it is a dependency of the `column_values.between.condition` metric.  # noqa: E501 # FIXME CoP
    table_columns_metric, table_column_metrics_results = get_table_columns_metric(
        execution_engine=execution_engine
    )
    desired_metric.metric_dependencies = {"table.columns": table_columns_metric}

    results = execution_engine.resolve_metrics(metrics_to_resolve=(desired_metric,))
    metrics = {
        "unexpected_condition": results[desired_metric.id],
        "table.columns": table_column_metrics_results[table_columns_metric.id],
    }
    mp = MapMetricProvider()
    res = _sqlalchemy_column_map_condition_values(
        cls=mp,
        execution_engine=execution_engine,
        metric_domain_kwargs=metric_domain_kwargs,
        metric_value_kwargs=metric_value_kwargs,
        metrics=metrics,
    )
    # one value is out of range with row condition
    assert res == expected_result


@pytest.mark.spark
@pytest.mark.parametrize(
    "execution_engine_fixture_name, metric_domain_kwargs, expected_result",
    [
        (
            "spark_execution_engine_with_mini_taxi_loaded",
            {
                "column": "total_amount",
                "row_condition": 'col("pk_1")==0',
                "condition_parser": "great_expectations",
            },
            [],
        ),
        (
            "spark_execution_engine_with_mini_taxi_loaded",
            {
                "column": "total_amount",
                "row_condition": 'col("pk_1")>=0',
                "condition_parser": "great_expectations",
            },
            [14.8],
        ),
    ],
)
def test_spark_column_map_condition_values(
    execution_engine_fixture_name,
    metric_domain_kwargs,
    expected_result,
    request,
    spark_session,
):
    execution_engine = request.getfixturevalue(execution_engine_fixture_name)
    metric_value_kwargs = {
        "min_value": 0,
        "max_value": 10.0,
        "strict_min": False,
        "strict_max": False,
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        },
    }

    desired_metric = MetricConfiguration(
        metric_name="column_values.between.condition",
        metric_domain_kwargs=metric_domain_kwargs,
        metric_value_kwargs=metric_value_kwargs,
    )

    # table.columns metric has to be calculated and loaded first, because it is a dependency of the `column_values.between.condition` metric.  # noqa: E501 # FIXME CoP
    table_columns_metric, table_column_metrics_results = get_table_columns_metric(
        execution_engine=execution_engine
    )
    desired_metric.metric_dependencies = {"table.columns": table_columns_metric}
    results = execution_engine.resolve_metrics(metrics_to_resolve=(desired_metric,))
    metrics = {
        "unexpected_condition": results[desired_metric.id],
        "table.columns": table_column_metrics_results[table_columns_metric.id],
    }
    mp = MapMetricProvider()
    res = _spark_column_map_condition_values(
        cls=mp,
        execution_engine=execution_engine,
        metric_domain_kwargs=metric_domain_kwargs,
        metric_value_kwargs=metric_value_kwargs,
        metrics=metrics,
    )
    # one value is out of range with row condition
    assert res == expected_result



from __future__ import annotations

from great_expectations.compatibility.not_imported import NotImported

# GX optional imports
SQLALCHEMY_NOT_IMPORTED = NotImported(
    "sqlalchemy is not installed, please 'pip install sqlalchemy'"
)

try:
    import sqlalchemy
except ImportError:
    sqlalchemy = SQLALCHEMY_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from sqlalchemy.sql.selectable import Subquery
except (ImportError, AttributeError):
    Subquery = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy import engine
except ImportError:
    engine = SQLALCHEMY_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from sqlalchemy import dialects
except ImportError:
    dialects = SQLALCHEMY_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from sqlalchemy import inspect
except ImportError:
    inspect = SQLALCHEMY_NOT_IMPORTED

try:
    from sqlalchemy.dialects import sqlite
except (ImportError, AttributeError):
    sqlite = SQLALCHEMY_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from sqlalchemy.dialects import registry
except (ImportError, AttributeError):
    registry = SQLALCHEMY_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from sqlalchemy.engine import Dialect
except (ImportError, AttributeError):
    Dialect = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.engine import Inspector
except (ImportError, AttributeError):
    Inspector = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.engine import reflection
except (ImportError, AttributeError):
    reflection = SQLALCHEMY_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from sqlalchemy.engine import Connection
except (ImportError, AttributeError):
    Connection = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.engine import Engine
except (ImportError, AttributeError):
    Engine = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.engine import Row
except (ImportError, AttributeError):
    Row = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.engine.row import RowProxy
except (ImportError, AttributeError):
    RowProxy = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.engine.row import LegacyRow  # type: ignore[attr-defined] # FIXME CoP
except (ImportError, AttributeError):
    LegacyRow = SQLALCHEMY_NOT_IMPORTED

try:
    from sqlalchemy.engine.default import DefaultDialect
except (ImportError, AttributeError):
    DefaultDialect = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.engine import url
    from sqlalchemy.engine.url import URL
except (ImportError, AttributeError):
    url = SQLALCHEMY_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP
    URL = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.exc import DatabaseError
except (ImportError, AttributeError):
    DatabaseError = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.exc import IntegrityError
except (ImportError, AttributeError):
    IntegrityError = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.exc import NoSuchTableError
except (ImportError, AttributeError):
    NoSuchTableError = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.exc import OperationalError
except (ImportError, AttributeError):
    OperationalError = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.exc import ProgrammingError
except (ImportError, AttributeError):
    ProgrammingError = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.exc import SQLAlchemyError
except (ImportError, AttributeError):
    SQLAlchemyError = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.orm import declarative_base
except (ImportError, AttributeError):
    declarative_base = SQLALCHEMY_NOT_IMPORTED

try:
    from sqlalchemy.sql import functions
except (ImportError, AttributeError):
    functions = SQLALCHEMY_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from sqlalchemy.sql import Insert
except (ImportError, AttributeError):
    Insert = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.elements import literal
except (ImportError, AttributeError):
    literal = SQLALCHEMY_NOT_IMPORTED

try:
    from sqlalchemy.sql.elements import TextClause
except (ImportError, AttributeError):
    TextClause = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.elements import quoted_name
except (ImportError, AttributeError):
    quoted_name = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.elements import _anonymous_label
except (ImportError, AttributeError):
    _anonymous_label = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.elements import ColumnElement
except (ImportError, AttributeError):
    ColumnElement = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import Cast
except (ImportError, AttributeError):
    Cast = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import ColumnOperators
except (ImportError, AttributeError):
    ColumnOperators = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import CTE
except (ImportError, AttributeError):
    CTE = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import BinaryExpression
except (ImportError, AttributeError):
    BinaryExpression = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import BooleanClauseList
except (ImportError, AttributeError):
    BooleanClauseList = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import ColumnClause
except (ImportError, AttributeError):
    ColumnClause = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import ClauseElement
except (ImportError, AttributeError):
    ClauseElement = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import Label
except (ImportError, AttributeError):
    Label = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import Select
except (ImportError, AttributeError):
    Select = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql import Selectable
except (ImportError, AttributeError):
    Selectable = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import TableClause
except (ImportError, AttributeError):
    TableClause = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import TextualSelect
except (ImportError, AttributeError):
    TextualSelect = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.expression import WithinGroup
except (ImportError, AttributeError):
    WithinGroup = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql.operators import custom_op
except (ImportError, AttributeError):
    custom_op = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.engine.cursor import (  # type: ignore[attr-defined] # FIXME CoP
        LegacyCursorResult,
    )
except (ImportError, AttributeError):
    LegacyCursorResult = SQLALCHEMY_NOT_IMPORTED

try:
    from sqlalchemy.engine.cursor import CursorResult
except (ImportError, AttributeError):
    CursorResult = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.pool import StaticPool
except (ImportError, AttributeError):
    StaticPool = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy import Table
except (ImportError, AttributeError):
    Table = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP


try:
    from sqlalchemy import Column
except (ImportError, AttributeError):
    Column = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP


try:
    from sqlalchemy import MetaData
except (ImportError, AttributeError):
    MetaData = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP


try:
    from sqlalchemy import create_engine
except (ImportError, AttributeError):
    create_engine = SQLALCHEMY_NOT_IMPORTED


try:
    from sqlalchemy import insert
except (ImportError, AttributeError):
    insert = SQLALCHEMY_NOT_IMPORTED

try:
    __version__: str | None = sqlalchemy.__version__
except (ImportError, AttributeError):
    __version__ = None

try:
    from sqlalchemy.sql.type_api import TypeEngine
except (ImportError, AttributeError):
    TypeEngine = SQLALCHEMY_NOT_IMPORTED  # type: ignore[misc,assignment] # FIXME CoP

try:
    from sqlalchemy.sql import sqltypes
except (ImportError, AttributeError):
    sqltypes = SQLALCHEMY_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP



import logging
import os
import uuid
import warnings
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union, cast

import pandas as pd
from sqlalchemy.exc import ProgrammingError

import great_expectations.exceptions as gx_exceptions
from great_expectations.alias_types import PathStr
from great_expectations.compatibility import sqlalchemy
from great_expectations.compatibility.sqlalchemy import Engine, inspect
from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.compatibility.sqlalchemy_compatibility_wrappers import (
    add_dataframe_to_db,
)
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.data_context.store import (
    CheckpointStore,
    ConfigurationStore,
    Store,
    StoreBackend,
)
from great_expectations.data_context.types.base import BaseYamlConfig
from great_expectations.data_context.types.resource_identifiers import (
    ConfigurationIdentifier,
)
from great_expectations.datasource.fluent.sql_datasource import SQLDatasource
from great_expectations.execution_engine import SqlAlchemyExecutionEngine
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect

logger = logging.getLogger(__name__)
yaml_handler = YAMLHandler()

SQLAlchemyError = sqlalchemy.SQLAlchemyError


def safe_remove(path):
    if path is not None:
        try:
            os.remove(path)  # noqa: PTH107 # FIXME CoP
        except OSError as e:
            print(e)


def create_files_in_directory(
    directory: str, file_name_list: List[str], file_content_fn=lambda: "x,y\n1,2\n2,3"
):
    subdirectories = []
    for file_name in file_name_list:
        splits = file_name.split("/")
        for i in range(1, len(splits)):
            subdirectories.append(os.path.join(*splits[:i]))  # noqa: PTH118 # FIXME CoP
    subdirectories = set(subdirectories)

    for subdirectory in subdirectories:
        os.makedirs(  # noqa: PTH103 # FIXME CoP
            os.path.join(directory, subdirectory),  # noqa: PTH118 # FIXME CoP
            exist_ok=True,
        )

    for file_name in file_name_list:
        file_path = os.path.join(directory, file_name)  # noqa: PTH118 # FIXME CoP
        with open(file_path, "w") as f_:
            f_.write(file_content_fn())


def validate_uuid4(uuid_string: str) -> bool:
    """
    Validate that a UUID string is in fact a valid uuid4.
    Happily, the uuid module does the actual checking for us.
    It is vital that the 'version' kwarg be passed
    to the UUID() call, otherwise any 32-character
    hex string is considered valid.
    From https://gist.github.com/ShawnMilo/7777304

    Args:
        uuid_string: string to check whether it is a valid UUID or not

    Returns:
        True if uuid_string is a valid UUID or False if not
    """
    try:
        val = uuid.UUID(uuid_string, version=4)
    except ValueError:
        # If it's a value error, then the string
        # is not a valid hex code for a UUID.
        return False

    # If the uuid_string is a valid hex code,
    # but an invalid uuid4,
    # the UUID.__init__ will convert it to a
    # valid uuid4. This is bad for validation purposes.

    return val.hex == uuid_string.replace("-", "")


def get_sqlite_temp_table_names(execution_engine: SqlAlchemyExecutionEngine):
    statement = sa.text("SELECT name FROM sqlite_temp_master")

    rows = execution_engine.execute_query(statement).fetchall()
    return {row[0] for row in rows}


def get_sqlite_table_names(execution_engine: SqlAlchemyExecutionEngine):
    statement = sa.text("SELECT name FROM sqlite_master")

    rows = execution_engine.execute_query(statement).fetchall()

    return {row[0] for row in rows}


def get_sqlite_temp_table_names_from_engine(engine: Engine):
    statement = sa.text("SELECT name FROM sqlite_temp_master")

    with engine.connect() as connection:
        rows = connection.execute(statement).fetchall()
    return {row[0] for row in rows}


def build_tuple_filesystem_store_backend(
    base_directory: str,
    *,
    module_name: str = "great_expectations.data_context.store",
    class_name: str = "TupleFilesystemStoreBackend",
    **kwargs,
) -> StoreBackend:
    logger.debug(
        f"""Starting data_context/store/util.py#build_tuple_filesystem_store_backend using base_directory:
"{base_directory}"""  # noqa: E501 # FIXME CoP
    )
    store_backend_config: dict = {
        "module_name": module_name,
        "class_name": class_name,
        "base_directory": base_directory,
    }
    store_backend_config.update(**kwargs)
    return Store.build_store_from_config(
        config=store_backend_config,
        module_name=module_name,
        runtime_environment=None,
    )


def save_config_to_filesystem(
    configuration_store_class_name: str,
    configuration_store_module_name: str,
    store_name: str,
    base_directory: str,
    configuration_key: str,
    configuration: BaseYamlConfig,
):
    store_config: dict = {"base_directory": base_directory}
    store_backend_obj: StoreBackend = build_tuple_filesystem_store_backend(**store_config)
    save_config_to_store_backend(
        class_name=configuration_store_class_name,
        module_name=configuration_store_module_name,
        store_name=store_name,
        store_backend=store_backend_obj,
        configuration_key=configuration_key,
        configuration=configuration,
    )


def load_config_from_filesystem(
    configuration_store_class_name: str,
    configuration_store_module_name: str,
    store_name: str,
    base_directory: str,
    configuration_key: str,
) -> BaseYamlConfig:
    store_config: dict = {"base_directory": base_directory}
    store_backend_obj: StoreBackend = build_tuple_filesystem_store_backend(**store_config)
    return load_config_from_store_backend(
        class_name=configuration_store_class_name,
        module_name=configuration_store_module_name,
        store_name=store_name,
        store_backend=store_backend_obj,
        configuration_key=configuration_key,
    )


def build_configuration_store(
    class_name: str,
    store_name: str,
    store_backend: Union[StoreBackend, dict],
    *,
    module_name: str = "great_expectations.data_context.store",
    overwrite_existing: bool = False,
    **kwargs,
) -> ConfigurationStore:
    logger.debug(
        f"Starting data_context/store/util.py#build_configuration_store for store_name {store_name}"
    )

    if store_backend is not None and isinstance(store_backend, StoreBackend):
        store_backend = store_backend.config
    elif not isinstance(store_backend, dict):
        raise gx_exceptions.DataContextError(
            "Invalid configuration: A store_backend needs to be a dictionary or inherit from the StoreBackend class."  # noqa: E501 # FIXME CoP
        )

    store_backend.update(**kwargs)

    store_config: dict = {
        "store_name": store_name,
        "module_name": module_name,
        "class_name": class_name,
        "overwrite_existing": overwrite_existing,
        "store_backend": store_backend,
    }
    configuration_store: ConfigurationStore = Store.build_store_from_config(  # type: ignore[assignment] # FIXME CoP
        config=store_config,
        module_name=module_name,
        runtime_environment=None,
    )
    return configuration_store


def delete_config_from_store_backend(
    class_name: str,
    module_name: str,
    store_name: str,
    store_backend: Union[StoreBackend, dict],
    configuration_key: str,
) -> None:
    config_store: ConfigurationStore = build_configuration_store(
        class_name=class_name,
        module_name=module_name,
        store_name=store_name,
        store_backend=store_backend,
        overwrite_existing=True,
    )
    key = ConfigurationIdentifier(
        configuration_key=configuration_key,
    )
    config_store.remove_key(key=key)


def delete_checkpoint_config_from_store_backend(
    store_name: str,
    store_backend: Union[StoreBackend, dict],
    checkpoint_name: str,
) -> None:
    config_store: CheckpointStore = build_checkpoint_store_using_store_backend(
        store_name=store_name,
        store_backend=store_backend,
    )
    key = ConfigurationIdentifier(
        configuration_key=checkpoint_name,
    )
    config_store.remove_key(key=key)


def build_checkpoint_store_using_store_backend(
    store_name: str,
    store_backend: Union[StoreBackend, dict],
    overwrite_existing: bool = False,
) -> CheckpointStore:
    return cast(
        "CheckpointStore",
        build_configuration_store(
            class_name="CheckpointStore",
            module_name="great_expectations.data_context.store",
            store_name=store_name,
            store_backend=store_backend,
            overwrite_existing=overwrite_existing,
        ),
    )


def save_config_to_store_backend(
    class_name: str,
    module_name: str,
    store_name: str,
    store_backend: Union[StoreBackend, dict],
    configuration_key: str,
    configuration: BaseYamlConfig,
) -> None:
    config_store: ConfigurationStore = build_configuration_store(
        class_name=class_name,
        module_name=module_name,
        store_name=store_name,
        store_backend=store_backend,
        overwrite_existing=True,
    )
    key = ConfigurationIdentifier(
        configuration_key=configuration_key,
    )
    config_store.set(key=key, value=configuration)


def load_config_from_store_backend(
    class_name: str,
    module_name: str,
    store_name: str,
    store_backend: Union[StoreBackend, dict],
    configuration_key: str,
) -> BaseYamlConfig:
    config_store: ConfigurationStore = build_configuration_store(
        class_name=class_name,
        module_name=module_name,
        store_name=store_name,
        store_backend=store_backend,
        overwrite_existing=False,
    )
    key = ConfigurationIdentifier(
        configuration_key=configuration_key,
    )
    return config_store.get(key=key)  # type: ignore[return-value] # FIXME CoP


def delete_config_from_filesystem(
    configuration_store_class_name: str,
    configuration_store_module_name: str,
    store_name: str,
    base_directory: str,
    configuration_key: str,
):
    store_config: dict = {"base_directory": base_directory}
    store_backend_obj: StoreBackend = build_tuple_filesystem_store_backend(**store_config)
    delete_config_from_store_backend(
        class_name=configuration_store_class_name,
        module_name=configuration_store_module_name,
        store_name=store_name,
        store_backend=store_backend_obj,
        configuration_key=configuration_key,
    )


# noinspection PyPep8Naming
def get_snowflake_connection_url() -> str:
    """Get snowflake connection url from environment variables.

    Returns:
        String of the snowflake connection URL.
    """
    sfUser = os.environ.get("SNOWFLAKE_USER")
    sfPswd = os.environ.get("SNOWFLAKE_PW")
    sfAccount = os.environ.get("SNOWFLAKE_ACCOUNT")
    sfDatabase = os.environ.get("SNOWFLAKE_DATABASE")
    sfSchema = os.environ.get("SNOWFLAKE_SCHEMA")
    sfWarehouse = os.environ.get("SNOWFLAKE_WAREHOUSE")
    sfRole = os.environ.get("SNOWFLAKE_ROLE") or "PUBLIC"

    return f"snowflake://{sfUser}:{sfPswd}@{sfAccount}/{sfDatabase}/{sfSchema}?warehouse={sfWarehouse}&role={sfRole}"


def get_redshift_connection_url() -> str:
    """Get Amazon Redshift connection url from environment variables.

    Returns:
        String of the Amazon Redshift connection URL.
    """
    host = os.environ.get("REDSHIFT_HOST")
    port = os.environ.get("REDSHIFT_PORT")
    user = os.environ.get("REDSHIFT_USERNAME")
    pswd = os.environ.get("REDSHIFT_PASSWORD")
    db = os.environ.get("REDSHIFT_DATABASE")
    ssl = os.environ.get("REDSHIFT_SSLMODE")

    return f"redshift+psycopg2://{user}:{pswd}@{host}:{port}/{db}?sslmode={ssl}"


def get_bigquery_table_prefix() -> str:
    """Get table_prefix that will be used by the BigQuery client in loading a BigQuery table from DataFrame.

    Reference link
        https://cloud.google.com/bigquery/docs/samples/bigquery-load-table-dataframe

    Returns:
        String of table prefix, which is the gcp_project and dataset concatenated by a "."
    """  # noqa: E501 # FIXME CoP
    gcp_project = os.environ.get("GE_TEST_GCP_PROJECT")
    if not gcp_project:
        raise ValueError(
            "Environment Variable GE_TEST_GCP_PROJECT is required to run BigQuery integration tests"
        )
    bigquery_dataset = os.environ.get("GE_TEST_BIGQUERY_DATASET", "test_ci")
    return f"""{gcp_project}.{bigquery_dataset}"""


def get_bigquery_connection_url() -> str:
    """Get bigquery connection url from environment variables.

    Note: dataset defaults to "demo" if not set.

    Returns:
        String of the bigquery connection url.
    """
    gcp_project = os.environ.get("GE_TEST_GCP_PROJECT")
    if not gcp_project:
        raise ValueError(
            "Environment Variable GE_TEST_GCP_PROJECT is required to run BigQuery integration tests"
        )
    bigquery_dataset = os.environ.get("GE_TEST_BIGQUERY_DATASET", "demo")

    return f"bigquery://{gcp_project}/{bigquery_dataset}"


@dataclass
class LoadedTable:
    """Output of loading a table via load_data_into_test_database."""

    table_name: str
    inserted_dataframe: pd.DataFrame


def load_and_concatenate_csvs(
    csv_paths: List[str],
    load_full_dataset: bool = False,
    convert_column_names_to_datetime: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Utility method that is used in loading test data into a pandas dataframe.

    It includes several parameters used to describe the output data.

    Args:
        csv_paths: list of paths of csvs to write, can be a single path. These should all be the same shape.
        load_full_dataset: if False, load only the first 10 rows.
        convert_column_names_to_datetime: List of column names to convert to datetime before writing to db.

    Returns:
        A pandas dataframe concatenating data loaded from all csvs.
    """  # noqa: E501 # FIXME CoP

    if convert_column_names_to_datetime is None:
        convert_column_names_to_datetime = []

    import pandas as pd

    dfs: List[pd.DataFrame] = []
    for csv_path in csv_paths:
        df = pd.read_csv(csv_path)
        convert_string_columns_to_datetime(
            df=df, column_names_to_convert=convert_column_names_to_datetime
        )
        if not load_full_dataset:
            # Improving test performance by only loading the first 10 rows of our test data into the db  # noqa: E501 # FIXME CoP
            df = df.head(10)

        dfs.append(df)

    all_dfs_concatenated: pd.DataFrame = pd.concat(dfs)

    return all_dfs_concatenated


def convert_string_columns_to_datetime(
    df: pd.DataFrame, column_names_to_convert: Optional[List[str]] = None
) -> None:
    """
    Converts specified columns (e.g., "pickup_datetime" and "dropoff_datetime") to datetime column type.
    Side-effect: Passed DataFrame is modified (in-place).
    """  # noqa: E501 # FIXME CoP
    if column_names_to_convert is None:
        column_names_to_convert = []

    column_name_to_convert: str
    for column_name_to_convert in column_names_to_convert:
        df[column_name_to_convert] = pd.to_datetime(df[column_name_to_convert])


def load_data_into_test_database(  # noqa: C901, PLR0912, PLR0915 # FIXME CoP
    table_name: str,
    connection_string: str,
    schema_name: Optional[str] = None,
    csv_path: Optional[str] = None,
    csv_paths: Optional[List[str]] = None,
    load_full_dataset: bool = False,
    convert_colnames_to_datetime: Optional[List[str]] = None,
    random_table_suffix: bool = False,
    to_sql_method: Optional[str] = None,
    drop_existing_table: bool = True,
) -> LoadedTable:
    """Utility method that is used in loading test data into databases that can be accessed through SqlAlchemy.

    This includes local Dockerized DBs like postgres, but also cloud-dbs like BigQuery and Redshift.

    Args:
        table_name: name of the table to write to.
        connection_string: used to connect to the database.
        schema_name: optional name of schema
        csv_path: path of a single csv to write.
        csv_paths: list of paths of csvs to write.
        load_full_dataset: if False, load only the first 10 rows.
        convert_colnames_to_datetime: List of column names to convert to datetime before writing to db.
        random_table_suffix: If true, add 8 random characters to the table suffix and remove other tables with the
            same prefix.
        to_sql_method: Method to pass to method param of pd.to_sql()
        drop_existing_table: boolean value. If set to false, will append to existing table
    Returns:
        LoadedTable which for convenience, contains the pandas dataframe that was used to load the data.
    """  # noqa: E501 # FIXME CoP
    if csv_path and csv_paths:
        csv_paths.append(csv_path)
    elif csv_path and not csv_paths:
        csv_paths = [csv_path]

    all_dfs_concatenated: pd.DataFrame = load_and_concatenate_csvs(
        csv_paths, load_full_dataset, convert_colnames_to_datetime
    )

    if random_table_suffix:
        table_name: str = f"{table_name}_{str(uuid.uuid4())[:8]}"

    return_value: LoadedTable = LoadedTable(
        table_name=table_name, inserted_dataframe=all_dfs_concatenated
    )
    connection = None
    if sa:
        engine = sa.create_engine(connection_string)
    else:
        logger.debug(
            "Attempting to load data in to tests SqlAlchemy database, but unable to load SqlAlchemy context; "  # noqa: E501 # FIXME CoP
            "install optional sqlalchemy dependency for support."
        )
        return return_value

    if engine.dialect.name.lower().startswith("mysql"):
        # Don't attempt to DROP TABLE IF EXISTS on a table that doesn't exist in mysql because it will error  # noqa: E501 # FIXME CoP
        inspector = inspect(engine)
        db_name = connection_string.split("/")[-1]
        table_names = [name for name in inspector.get_table_names(schema=db_name)]
        drop_existing_table = table_name in table_names

    if engine.dialect.name.lower() == "bigquery":
        # bigquery is handled in a special way
        load_data_into_test_bigquery_database_with_bigquery_client(
            dataframe=all_dfs_concatenated, table_name=table_name
        )
        return return_value
    elif engine.dialect.name.lower() == "awsathena":
        try:
            connection = engine.connect()
            load_dataframe_into_test_athena_database_as_table(
                df=all_dfs_concatenated,
                table_name=table_name,
                connection=connection,
            )
            return return_value
        except SQLAlchemyError:
            error_message: str = """Docs integration tests encountered an error while loading test-data into test-database."""  # noqa: E501 # FIXME CoP
            logger.error(error_message)  # noqa: TRY400 # FIXME CoP
            raise gx_exceptions.DatabaseConnectionError(error_message)
            # Normally we would call `raise` to re-raise the SqlAlchemyError but we don't to make sure that  # noqa: E501 # FIXME CoP
            # sensitive information does not make it into our CI logs.
        finally:
            connection.close()
            engine.dispose()
    else:
        try:
            if drop_existing_table:
                print(f"Dropping table {table_name}")
                with engine.begin() as connection:
                    connection.execute(sa.text(f"DROP TABLE IF EXISTS {table_name}"))
                print(f"Creating table {table_name} and adding data from {csv_paths}")
            else:
                print(f"Adding to existing table {table_name} and adding data from {csv_paths}")

            with engine.connect() as connection:
                add_dataframe_to_db(
                    df=all_dfs_concatenated,
                    name=table_name,
                    con=connection,
                    schema=schema_name,
                    index=False,
                    if_exists="append",
                    method=to_sql_method,
                )
            return return_value
        except SQLAlchemyError:
            error_message: str = """Docs integration tests encountered an error while loading test-data into test-database."""  # noqa: E501 # FIXME CoP
            logger.error(error_message)  # noqa: TRY400 # FIXME CoP
            raise gx_exceptions.DatabaseConnectionError(error_message)
            # Normally we would call `raise` to re-raise the SqlAlchemyError but we don't to make sure that  # noqa: E501 # FIXME CoP
            # sensitive information does not make it into our CI logs.
        finally:
            if connection:
                connection.close()
            if engine:
                engine.dispose()


def load_data_into_test_bigquery_database_with_bigquery_client(
    dataframe: pd.DataFrame, table_name: str
) -> None:
    """
    Loads dataframe into bigquery table using BigQuery client. Follows pattern specified in the GCP documentation here:
        - https://cloud.google.com/bigquery/docs/samples/bigquery-load-table-dataframe
    Args:
        dataframe (pd.DataFrame): DataFrame to load
        table_name (str): table to load DataFrame to. Prefix containing project and dataset are loaded
                        by helper function.
    """  # noqa: E501 # FIXME CoP
    prefix: str = get_bigquery_table_prefix()
    table_id: str = f"""{prefix}.{table_name}"""
    from google.cloud import bigquery

    gcp_project: Optional[str] = os.environ.get("GE_TEST_GCP_PROJECT")
    if not gcp_project:
        raise ValueError(
            "Environment Variable GE_TEST_GCP_PROJECT is required to run BigQuery integration tests"
        )
    client: bigquery.Client = bigquery.Client(project=gcp_project)
    job: bigquery.LoadJob = client.load_table_from_dataframe(
        dataframe, table_id
    )  # Make an API request.
    load_job = job.result()  # Wait for the job to complete
    if load_job.errors or load_job.error_result:
        raise gx_exceptions.DatabaseConnectionError(
            "Unable to load data into BigQuery. Load job errors: "
            + str(load_job.errors)
            + " Load job error result: "
            + str(load_job.error_result)
        )


def load_dataframe_into_test_athena_database_as_table(
    df: pd.DataFrame,
    table_name: str,
    connection,
    data_location_bucket: Optional[str] = None,
    data_location: Optional[str] = None,
) -> None:
    """

    Args:
        df: dataframe containing data.
        table_name: name of table to write.
        connection: connection to database.
        data_location_bucket: name of bucket where data is located.
        data_location: path to data from bucket without leading / e.g.
            "data/stuff/" in path "s3://my-bucket/data/stuff/"

    Returns:
        None
    """

    from pyathena.pandas.util import to_sql

    if not data_location_bucket:
        data_location_bucket = os.getenv("ATHENA_DATA_BUCKET")
    if not data_location:
        data_location = "data/ten_trips_from_each_month/"
    location: str = f"s3://{data_location_bucket}/{data_location}"
    to_sql(
        df=df,
        name=table_name,
        conn=connection,
        location=location,
        if_exists="replace",
    )


def clean_up_tables_with_prefix(connection_string: str, table_prefix: str) -> List[str]:
    """Drop all tables starting with the provided table_prefix.
    Note: Uses private method InferredAssetSqlDataConnector._introspect_db()
    to get the table names to not duplicate code, but should be refactored in the
    future to not use a private method.

    Args:
        connection_string: To connect to the database.
        table_prefix: First characters of the tables you want to remove.

    Returns:
        List of deleted tables.
    """
    execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(
        connection_string=connection_string
    )
    introspection_output = introspect_db(execution_engine=execution_engine)

    tables_to_drop: List[str] = []
    tables_dropped: List[str] = []

    for table in introspection_output:
        if table["table_name"].startswith(table_prefix):
            tables_to_drop.append(table["table_name"])

    for table_name in tables_to_drop:
        print(f"Dropping table {table_name}")
        execution_engine.execute_query_in_transaction(sa.text(f"DROP TABLE IF EXISTS {table_name}"))
        tables_dropped.append(table_name)

    tables_skipped: List[str] = list(set(tables_to_drop) - set(tables_dropped))
    if len(tables_skipped) > 0:
        warnings.warn(f"Warning: Tables skipped: {tables_skipped}")

    return tables_dropped


def introspect_db(  # noqa: C901, PLR0912 # FIXME CoP
    execution_engine: SqlAlchemyExecutionEngine,
    schema_name: Union[str, None] = None,
    ignore_information_schemas_and_system_tables: bool = True,
    information_schemas: Optional[List[str]] = None,
    system_tables: Optional[List[str]] = None,
    include_views=True,
) -> List[Dict[str, str]]:
    # This code was broken out from the InferredAssetSqlDataConnector when it was removed
    if information_schemas is None:
        information_schemas = [
            "INFORMATION_SCHEMA",  # snowflake, mssql, mysql, oracle
            "information_schema",  # postgres, redshift, mysql
            "performance_schema",  # mysql
            "sys",  # mysql
            "mysql",  # mysql
        ]

    if system_tables is None:
        system_tables = ["sqlite_master"]  # sqlite

    engine: sqlalchemy.Engine = execution_engine.engine
    inspector: sqlalchemy.Inspector = sa.inspect(engine)

    selected_schema_name = schema_name

    tables: List[Dict[str, str]] = []
    all_schema_names: List[str] = inspector.get_schema_names()
    for schema in all_schema_names:
        if ignore_information_schemas_and_system_tables and schema_name in information_schemas:
            continue

        if selected_schema_name is not None and schema_name != selected_schema_name:
            continue

        try:
            table_names: List[str] = inspector.get_table_names(schema=schema)
        except ProgrammingError:
            # Likely another test already cleaned up this schema.
            # TODO: Make tests only clean up after themselves
            continue
        for table_name in table_names:
            if ignore_information_schemas_and_system_tables and (table_name in system_tables):
                continue

            tables.append(
                {
                    "schema_name": schema,
                    "table_name": table_name,
                    "type": "table",
                }
            )

        # Note Abe 20201112: This logic is currently untested.
        if include_views:
            # Note: this is not implemented for bigquery
            try:
                view_names = inspector.get_view_names(schema=schema)
            except NotImplementedError:
                # Not implemented by Athena dialect
                pass
            else:
                for view_name in view_names:
                    if ignore_information_schemas_and_system_tables and (
                        view_name in system_tables
                    ):
                        continue

                    tables.append(
                        {
                            "schema_name": schema,
                            "table_name": view_name,
                            "type": "view",
                        }
                    )

    # SQLAlchemy's introspection does not list "external tables" in Redshift Spectrum (tables whose data is stored on S3).  # noqa: E501 # FIXME CoP
    # The following code fetches the names of external schemas and tables from a special table
    # 'svv_external_tables'.
    try:
        if engine.dialect.name.lower() == GXSqlDialect.REDSHIFT:
            # noinspection SqlDialectInspection,SqlNoDataSourceInspection
            result = execution_engine.execute_query(
                sa.text("select schemaname, tablename from svv_external_tables")
            ).fetchall()
            for row in result:
                tables.append(
                    {
                        "schema_name": row[0],
                        "table_name": row[1],
                        "type": "table",
                    }
                )
    except Exception as e:
        # Our testing shows that 'svv_external_tables' table is present in all Redshift clusters. This means that this  # noqa: E501 # FIXME CoP
        # exception is highly unlikely to fire.
        if "UndefinedTable" not in str(e):
            raise e  # noqa: TRY201 # FIXME CoP

    return tables


def check_athena_table_count(
    connection_string: str, db_name: str, expected_table_count: int
) -> bool:
    """
    Helper function used by awsathena integration test. Checks whether expected number of tables exist in database
    """  # noqa: E501 # FIXME CoP
    if sa:
        engine = sa.create_engine(connection_string)
    else:
        logger.debug(
            "Attempting to perform test on AWSAthena database, but unable to load SqlAlchemy context; "  # noqa: E501 # FIXME CoP
            "install optional sqlalchemy dependency for support."
        )
        return False

    connection = None
    try:
        connection = engine.connect()
        result = connection.execute(sa.text(f"SHOW TABLES in {db_name}")).fetchall()
        return len(result) == expected_table_count
    except SQLAlchemyError:
        error_message: str = """Docs integration tests encountered an error while loading test-data into test-database."""  # noqa: E501 # FIXME CoP
        logger.error(error_message)  # noqa: TRY400 # FIXME CoP
        raise gx_exceptions.DatabaseConnectionError(error_message)
        # Normally we would call `raise` to re-raise the SqlAlchemyError but we don't to make sure that  # noqa: E501 # FIXME CoP
        # sensitive information does not make it into our CI logs.
    finally:
        connection.close()
        engine.dispose()


def clean_athena_db(connection_string: str, db_name: str, table_to_keep: str) -> None:
    """
    Helper function used by awsathena integration test. Cleans up "temp" tables that were created.
    """
    if sa:
        engine = sa.create_engine(connection_string)
    else:
        logger.debug(
            "Attempting to perform test on AWSAthena database, but unable to load SqlAlchemy context; "  # noqa: E501 # FIXME CoP
            "install optional sqlalchemy dependency for support."
        )
        return

    connection = None
    try:
        connection = engine.connect()
        result = connection.execute(sa.text(f"SHOW TABLES in {db_name}")).fetchall()
        for table_tuple in result:
            table = table_tuple[0]
            if table != table_to_keep:
                connection.execute(sa.text(f"DROP TABLE `{table}`;"))
    finally:
        connection.close()
        engine.dispose()


def get_default_postgres_url() -> str:
    """Get connection string to postgres container
    Returns:
        String of default connection to Docker container
    """
    return "postgresql+psycopg2://postgres:@localhost/test_ci"


def get_default_mysql_url() -> str:
    """Get connection string to mysql container
    Returns:
        String of default connection to Docker container
    """
    return "mysql+pymysql://root@127.0.0.1/test_ci"


def get_default_trino_url() -> str:
    """Get connection string to trino container
    Returns:
        String of default connection to Docker container
    """
    return "trino://test@localhost:8088/memory/schema"


def get_default_mssql_url() -> str:
    """Get connection string to mssql container
    Returns:
        String of default connection to Docker container
    """
    return (
        "mssql+pyodbc://sa:ReallyStrongPwd1234%^&*@127.0.0.1:1433/test_ci"
        "?driver=ODBC Driver 18 for SQL Server&charset=utf8"
        "&autocommit=true&TrustServerCertificate=yes"
    )


def get_awsathena_db_name(db_name_env_var: str = "ATHENA_DB_NAME") -> str:
    """Get awsathena database name from environment variables.

    Returns:
        String of the awsathena database name.
    """
    athena_db_name: str = os.getenv(db_name_env_var)
    if not athena_db_name:
        raise ValueError(
            f"Environment Variable {db_name_env_var} is required to run integration tests against AWS Athena"  # noqa: E501 # FIXME CoP
        )
    return athena_db_name


def get_awsathena_connection_url(db_name_env_var: str = "ATHENA_DB_NAME") -> str:
    """Get awsathena connection url from environment variables.

    Returns:
        String of the awsathena connection url.
    """
    ATHENA_DB_NAME: str = get_awsathena_db_name(db_name_env_var)
    ATHENA_STAGING_S3: Optional[str] = os.getenv("ATHENA_STAGING_S3")
    if not ATHENA_STAGING_S3:
        raise ValueError(
            "Environment Variable ATHENA_STAGING_S3 is required to run integration tests against AWS Athena"  # noqa: E501 # FIXME CoP
        )

    return f"awsathena+rest://@athena.us-east-1.amazonaws.com/{ATHENA_DB_NAME}?s3_staging_dir={ATHENA_STAGING_S3}"


def get_connection_string_and_dialect(
    athena_db_name_env_var: str = "ATHENA_DB_NAME",
) -> Tuple[str, str]:
    with open("./connection_string.yml") as f:
        db_config: dict = yaml_handler.load(f)

    dialect: str = db_config["dialect"]
    if dialect == "snowflake":
        connection_string: str = get_snowflake_connection_url()
    elif dialect == "redshift":
        connection_string: str = get_redshift_connection_url()
    elif dialect == "bigquery":
        connection_string: str = get_bigquery_connection_url()
    elif dialect == "awsathena":
        connection_string: str = get_awsathena_connection_url(athena_db_name_env_var)
    else:
        connection_string: str = db_config["connection_string"]

    return dialect, connection_string


def add_datasource(
    context: AbstractDataContext, *, name: str, connection_string: str
) -> SQLDatasource:
    """Add a datasource to the context based on the dialect from config file.

    Needed because context.data_sources.add_sql is prohibitted when
    more specific methods are available.
    """
    with open("./connection_string.yml") as f:
        db_config: dict = yaml_handler.load(f)

    dialect: str = db_config["dialect"]
    if dialect == "snowflake":
        return context.data_sources.add_snowflake(name=name, connection_string=connection_string)
    elif dialect == "postgres":
        return context.data_sources.add_postgres(name=name, connection_string=connection_string)
    elif dialect == "redshift":
        return context.data_sources.add_redshift(name=name, connection_string=connection_string)
    elif dialect == "bigquery":
        return context.data_sources.add_bigquery(name=name, connection_string=connection_string)
    else:
        return context.data_sources.add_sql(name=name, connection_string=connection_string)


@contextmanager
def working_directory(directory: PathStr):
    """
    Resets working directory to cwd() after changing to 'directory' passed in as parameter.
    Reference:
    https://stackoverflow.com/questions/431684/equivalent-of-shell-cd-command-to-change-the-working-directory/431747#431747
    """
    owd = os.getcwd()  # noqa: PTH109 # FIXME CoP
    try:
        os.chdir(directory)
        yield directory
    finally:
        os.chdir(owd)



from __future__ import annotations

import copy
import logging
import pathlib
from typing import TYPE_CHECKING

import pytest

from great_expectations.datasource.fluent import TestConnectionError
from great_expectations.datasource.fluent.spark_datasource import (
    DataFrameAsset,
    SparkConfig,
)
from great_expectations.exceptions.exceptions import BuildBatchRequestError
from great_expectations.execution_engine.sparkdf_batch_data import SparkDFBatchData
from great_expectations.util import is_candidate_subset_of_target

if TYPE_CHECKING:
    import pandas as pd

    from great_expectations.data_context import AbstractDataContext


logger = logging.getLogger(__file__)


@pytest.fixture
def valid_file_path(csv_path: pathlib.Path) -> pathlib.Path:
    return csv_path / "yellow_tripdata_sample_2018-03.csv"


@pytest.mark.spark
def test_dataframe_asset(
    empty_data_context: AbstractDataContext,
    spark_session,
    spark_df_from_pandas_df,
    test_df_pandas,
):
    datasource = empty_data_context.data_sources.add_spark(name="my_spark_datasource")

    pandas_df = test_df_pandas
    spark_df = spark_df_from_pandas_df(spark_session, pandas_df)

    dataframe_asset = datasource.add_dataframe_asset(
        name="my_dataframe_asset",
    )
    assert isinstance(dataframe_asset, DataFrameAsset)
    assert dataframe_asset.name == "my_dataframe_asset"
    assert len(datasource.assets) == 1

    datasource.add_dataframe_asset(
        name="my_second_dataframe_asset",
    )
    assert len(datasource.assets) == 2

    for i, asset in enumerate(datasource.assets):
        bd = asset.add_batch_definition_whole_dataframe(name=f"bd_{i}")
        batch = bd.get_batch(batch_parameters={"dataframe": spark_df})
        assert isinstance(batch.data, SparkDFBatchData)
        assert batch.data.dataframe.toPandas().equals(pandas_df)


@pytest.mark.spark
def test_spark_data_asset_batch_metadata(
    empty_data_context: AbstractDataContext,
    valid_file_path: pathlib.Path,
    test_df_pandas: pd.DataFrame,
    spark_session,
    spark_df_from_pandas_df,
):
    my_config_variables = {"pipeline_filename": __file__}
    empty_data_context.config_variables.update(my_config_variables)

    spark_df = spark_df_from_pandas_df(spark_session, test_df_pandas)

    spark_datasource = empty_data_context.data_sources.add_spark("my_spark_datasource")

    batch_metadata = {
        "no_curly_pipeline_filename": "$pipeline_filename",
        "curly_pipeline_filename": "${pipeline_filename}",
        "pipeline_step": "transform_3",
    }

    dataframe_asset = spark_datasource.add_dataframe_asset(
        name="my_dataframe_asset",
        batch_metadata=batch_metadata,
    )
    assert dataframe_asset.batch_metadata == batch_metadata

    batch = dataframe_asset.get_batch(
        dataframe_asset.build_batch_request(options={"dataframe": spark_df})
    )
    substituted_batch_metadata = copy.deepcopy(batch_metadata)
    substituted_batch_metadata.update(
        {
            "no_curly_pipeline_filename": __file__,
            "curly_pipeline_filename": __file__,
        }
    )
    assert batch.metadata == substituted_batch_metadata


@pytest.mark.spark
@pytest.mark.parametrize("persist", [True, False])
def test_spark_config_passed_to_execution_engine(
    empty_data_context: AbstractDataContext,
    persist,
    spark_session,
):
    spark_config: SparkConfig = {
        "spark.app.name": "gx_spark_fluent_datasource_test",
        "spark.default.parallelism": 4,
        "spark.master": "local[*]",
    }
    datasource = empty_data_context.data_sources.add_spark(
        name="my_spark_datasource",
        spark_config=spark_config,
        persist=persist,
    )
    execution_engine_spark_config = datasource.get_execution_engine().config["spark_config"]
    assert is_candidate_subset_of_target(
        candidate=spark_config,
        target=execution_engine_spark_config,
    )


@pytest.mark.spark
def test_build_batch_request_raises_if_missing_dataframe(
    empty_data_context: AbstractDataContext,
    spark_session,
):
    dataframe_asset = empty_data_context.data_sources.add_spark(
        name="my_spark_datasource"
    ).add_dataframe_asset(name="my_dataframe_asset")

    with pytest.raises(BuildBatchRequestError) as e:
        dataframe_asset.build_batch_request()

    assert "options must contain exactly 1 key, 'dataframe'" in str(e.value)


@pytest.mark.spark
def test_unmodifiable_config_option_warning(
    empty_data_context: AbstractDataContext,
    spark_session,
):
    spark_config = {"spark.executor.memory": "700m"}
    with pytest.warns(RuntimeWarning):
        _ = empty_data_context.data_sources.add_spark(
            name="my_spark_datasource",
            spark_config=spark_config,  # type: ignore[arg-type] # FIXME CoP
        )


@pytest.mark.unit
def test_spark_test_connection(
    empty_data_context: AbstractDataContext,
):
    # no spark marker means pyspark is not installed when this is run
    with pytest.raises(TestConnectionError):
        _ = empty_data_context.data_sources.add_spark(name="my_spark_datasource")



from __future__ import annotations

import warnings

from great_expectations.compatibility.not_imported import NotImported

DOCSTRING_PARSER_NOT_IMPORTED = NotImported(
    "docstring_parser is not installed, please 'pip install docstring-parser'"
)
with warnings.catch_warnings():
    # Hide deprecation warnings emitted by compatibility dependencies, users have no control here.
    warnings.simplefilter("ignore", category=DeprecationWarning)
    try:
        import docstring_parser
    except ImportError:
        docstring_parser = DOCSTRING_PARSER_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

    try:
        from docstring_parser import DocstringStyle
    except ImportError:
        DocstringStyle = DOCSTRING_PARSER_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP
